# filestore: Dmedia hashing protocol and file layout
# Copyright (C) 2012 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Helper functions for creating a utilizing protocol test vectors.
"""

from os import path
import json
import tempfile
import shutil
from hashlib import md5
import binascii

from dbase32 import db32enc, db32dec, random_id

from .protocols import PERS_LEAF, PERS_ROOT, VERSION1
from . import FileStore, _create_filestore, ContentHash, _dumps


DATADIR = path.join(path.dirname(path.abspath(__file__)), 'data')


def tohex(data):
    return binascii.hexlify(data).decode('utf-8')


def load_data(name):
    filename = path.join(DATADIR, name + '.json')
    return json.load(open(filename, 'r'))


def build_leaves(leaf_size):
    return {
        'A': b'A',
        'B': b'B' * (leaf_size - 1),
        'C': b'C' * leaf_size,
    }


def write_test_vector_files(tmpdir, protocol=VERSION1):
    """
    Write standard test vector files into directory *tmpdir*.
    """
    leaves = build_leaves(protocol.leaf_size)
    for (key, data) in leaves.items():
        open(path.join(tmpdir, key), 'xb').write(data)
        fp = open(path.join(tmpdir, 'C' + key), 'xb')
        fp.write(leaves['C'])
        fp.write(data)
        fp.close()


def import_vector_files(fs):
    leaves = build_leaves(VERSION1.leaf_size)
    vectors = vectors = load_data('V1')
    result = []
    for (F, _id) in vectors['root_hashes'].items():
        file_size = 0
        leaf_hashes = bytearray()
        fp = open(fs.path(_id), 'xb')
        for (i, L) in enumerate(F):
            leaf_data = leaves[L]
            fp.write(leaf_data)
            file_size += len(leaf_data)
            leaf_hashes.extend(db32dec(vectors['leaf_hashes'][L][i]))
        fp.close()
        result.append(ContentHash(_id, file_size, bytes(leaf_hashes)))
    return tuple(
        sorted(result, key=lambda ch: ch.id)
    )


def build_vectors(protocol, encoder=db32enc):
    leaves = build_leaves(protocol.leaf_size)

    def hash_leaf(i, L):
        return encoder(protocol.hash_leaf(i, leaves[L]))

    def hash_root(name):
        leaf_hashes = b''
        file_size = 0
        for (i, L) in enumerate(name):
            data = leaves[L]
            leaf_hashes += protocol.hash_leaf(i, data)
            file_size += len(data)
        return encoder(protocol.hash_root(file_size, leaf_hashes))

    vectors = {
        'leaf_hashes': {},
        'root_hashes': {},
    }

    for L in leaves:
        vectors['leaf_hashes'][L] = [hash_leaf(i, L) for i in range(2)]

    for L in leaves:
        for name in (L, 'C' + L):
            vectors['root_hashes'][name] = hash_root(name)

    return vectors


def get_integers(leaf_size):
    return [
        0,
        1,
        leaf_size - 1,
        leaf_size,
        leaf_size + 1,
        2 * leaf_size - 1,
        2 * leaf_size,
    ]


def build_md5sums(leaf_size):
    integers = dict(
        (str(i), md5(str(i).encode('utf-8')).hexdigest())
        for i in get_integers(leaf_size)
    )

    leaves = build_leaves(leaf_size)
    C = leaves['C']
    files = {}
    for (key, data) in leaves.items():
        files[key] = md5(data).hexdigest()
        files['C' + key] = md5(C + data).hexdigest()

    return {
        'integers': integers,
        'files': files,
        'personalization': {
            'leaf': md5(PERS_LEAF).hexdigest(),
            'root': md5(PERS_ROOT).hexdigest(),
        }   
    }


class TempFileStore(FileStore):
    """
    A throw-away `FileStore` for experimenting and testing.

    When you create a `TempFileStore` instance, a random, temporary
    `FileStore.parentdir` is created for you.  For example:

    >>> fs = TempFileStore()

    Then when the `TempFileStore` instance is garbage collected, this
    directory is automatically deleted (along with any files contained
    therein).

    As with a `FileStore`, you can provide the `FileStore.id` and
    `FileStore.copies` attributes.  The `FileStore` itself ignores these
    values, but they are convenient for high-level software like Dmedia.

    For example:

    >>> fs = TempFileStore('39AY39AY39AY39AY39AY39AY', 2)
    >>> fs.id
    '39AY39AY39AY39AY39AY39AY'
    >>> fs.copies
    2

    """

    @classmethod
    def create(cls, store_id=None, copies=1, **kw):
        return cls(store_id, copies, **kw)

    def __init__(self, store_id=None, copies=1, **kw):
        parentdir = tempfile.mkdtemp(prefix='TempFileStore.')
        store_id = (random_id() if store_id is None else store_id)
        _create_filestore(parentdir, store_id, copies, **kw)
        super().__init__(parentdir, store_id)
        assert self.parentdir == parentdir

    def __del__(self):
        super().__del__()
        if path.isdir(self.parentdir):
            shutil.rmtree(self.parentdir)


if __name__ == '__main__':
    print(_dumps(build_vectors(VERSION1, db32enc)))
    print('')
    print(_dumps(build_vectors(VERSION1, tohex)))

